package user

import (
	"context"
	"crypto/md5"
	"encoding/hex"
	"errors"
	"fmt"

	"github.com/jackc/pgx/v5"
	"github.com/jackc/pgx/v5/pgconn"
	"github.com/jackc/pgx/v5/pgxpool"
)

type UserEntity struct {
	Id       int
	Username string
	Password string
}

type UsernameAlreadyExistsErr struct{}

func (e *UsernameAlreadyExistsErr) Error() string {
	return "username already exists error"
}

type UserNotFoundErr struct{}

func (e *UserNotFoundErr) Error() string {
	return "user not found error"
}

type UserService struct {
	db *pgxpool.Pool
}

func NewUserService(
	db *pgxpool.Pool,
) *UserService {
	return &UserService{
		db: db,
	}
}

func (s *UserService) AddUser(ctx context.Context, user *UserEntity) (*UserEntity, error) {
	query := `INSERT INTO users (username, password) VALUES (@username, @password) RETURNING id`
	args := pgx.NamedArgs{
		"username": user.Username,
		"password": hashPassword(user.Username, user.Password, "crab"),
	}
	if err := s.db.QueryRow(ctx, query, args).Scan(&user.Id); err != nil {
		var pgErr *pgconn.PgError
		if errors.As(err, &pgErr) {
			if pgErr.Code == "23505" && pgErr.ConstraintName == "users_username_key" { // unique_violation
				return nil, &UsernameAlreadyExistsErr{}
			}
		}
		return nil, fmt.Errorf("unable to insert row: %w", err)
	}
	return user, nil
}

func (s *UserService) Login(ctx context.Context, user *UserEntity) (*UserEntity, error) {
	query := `SELECT id FROM users WHERE username = @username AND password = @password`
	args := pgx.NamedArgs{
		"username": user.Username,
		"password": hashPassword(user.Username, user.Password, "crab"),
	}
	if err := s.db.QueryRow(ctx, query, args).Scan(&user.Id); err != nil {
		return nil, &UserNotFoundErr{}
	}
	return user, nil
}

func hashPassword(username string, password string, salt string) string {
	return getMD5Hash(username + password + salt)
}

func getMD5Hash(text string) string {
	hash := md5.Sum([]byte(text))
	return hex.EncodeToString(hash[:])
}
